import { ILogger, IPargv } from 'pargv';

export type BeforeRender = { (template?: ITemplate, done?: { () }) };

export type AfterRender = { (template?: ITemplate) };

export type NodeCallback = { (err?: any, data?: any) };

export type InjectCallback = { (line: string, data: string, done: { (line: string, data: string) }) };

export interface IMetadataComponent {
  name?: string;
  fullname?: string;
  group?: string;
  ext?: string;
  path?: string;
  from?: string; // From prop updated on each iteration of injectable.
}

export interface IMetadata {
  $component?: IMetadataComponent;   // the component name generated from command line.
  [key: string]: any;   // other metadata passed in template or on rendering.
}

export interface IPartials {
  [key: string]: string | ITemplate;
}

export interface IConfig {

  name?: string;                // name of template used ONLY when passing static templates.
  ext?: string;                 // extension to use when rendered.
  type?: string;                // type suffix to be added to the component name.
  appendType?: boolean;         // when true type name is added to the output filename.
  casing?: 'lower' | 'upper' | 'camel' | 'title' | 'capitalize';  // component name case.
  filenameCasing?: 'lower' | 'upper' | 'camel' | 'title' | 'capitalize';
  injects?: any[];
  rename?: string;              // renames the file.
  outputDir?: string;           // base output dir or use in outputDir in options.
  beforeRender?: BeforeRender;  // the callback before rendering/compiling template.
  afterRender?: AfterRender;    // the callback after rending the template to file.
  partials?: string | string[] | IPartials;  // object conatianing template's partials.
  metadata?: IMetadata;         // metadata for template rendering.
  noOutput?: boolean;           // when true won't write out to file.

}

export interface ITemplate extends IConfig {

  absolute?: boolean;            // when outputPath is statically provided is absolute.
  path?: string;                // the path to the template if not static defined template.
  pathNormalized?: string;      // the path without template extension.
  body?: string;                // the template body.
  outputPath?: string;          // the path to output the template to.
  outputRelative?: string;      // path relative to the config's base output dir.
  raw?: string;                 // a static template that has not been parsed.
  rendered?: string;            // the rendered string result.
  isPartial?: boolean;          // indicates the template is a partial.
  group?: string;               // template is part of a group being rendered.

}

export interface ITemplates {
  [key: string]: ITemplate;
}

export interface IComponent {
  name?: string;
  templates: string[];
  series: boolean;
}

export interface IComponents {
  [key: string]: IComponent;
}

export interface IGroupOptions extends IComponent {
  series: boolean;      // when true assumes static outputPaths passed in template configs.
}

export interface IRegisterConfig extends IConfig {
  template?: string;
}

export interface IRegister {
  partials(...args: any[]): IRegister;
  partials(partials: string | string[] | IPartials): IRegister;
  beforeRender(beforeRender: BeforeRender): IRegister;
  afterRender(afterRender: AfterRender): IRegister;
}

export interface IInject {
  filename: string;
  find: string | RegExp;
  insert: string | string[];
  strategy: 'before' | 'after' | 'first' | 'last' | 'replace';
  done: NodeCallback;
  relative: boolean;      // when true filename is relative to the output path.
}

export interface IMustrOptions {
  configDir?: string;     // the Muster dir containing templates and register.
  outputDir?: string;     // the output directory when templates are rendered.
  templateExt?: string;   // the extension for templates usually .tpl.
  autoRegister?: boolean; // when true templates are auto registered using defaults.
}

export interface IMustr {

  cwd: string;
  log: ILogger;
  registerPath: string;
  templatesPath: string;
  outputPath: string;
  templatesGlob: string[];
  templates: ITemplates;
  components: IComponents;
  options: IMustrOptions;

  init(force?: boolean): void;
  load(): IMustr;

  register(options: IRegisterConfig): IRegister;
  register(name: string | IRegisterConfig, template?: string | IRegisterConfig, options?: IRegisterConfig): IRegister;

  registerGroup(name: string, series: boolean);
  registerGroup(name: string, ...args: string[]): void;

  configure(name: string, options: IRegisterConfig): ITemplate;
  configure(name: string, output: string, group: string);
  configure(name: string, output?: string | IRegisterConfig, options?: IRegisterConfig | string): ITemplate;

  render(template: ITemplate, done?: NodeCallback): void;
  render(template: ITemplate, force?: boolean, done?: NodeCallback): void;
  render(template: ITemplate, options?: IRegisterConfig, done?: NodeCallback): void;
  render(template: ITemplate, options?: IRegisterConfig, force?: boolean, done?: NodeCallback): void;
  render(name: string | ITemplate, output?: string | IRegisterConfig | NodeCallback | boolean, options?: IRegisterConfig | NodeCallback | boolean, force?: boolean | NodeCallback, done?: NodeCallback, group?: string): void;

  inject(options: IInject, done?: NodeCallback): void;
  inject(filename: string | IInject,
    find: string | RegExp | NodeCallback,
    insert: string | string[],
    strategy: 'before' | 'after' | 'first' | 'last' | 'replace',
    done?: NodeCallback): void;
  transformCase(str: string, to: string): string;

}